/**************************************************************************

	nirc

	view.c

	Copyright (C) 1996-2002 by Tomoaki Nakashima. All rights reserved.
		http://www.nakka.com/
		nakka@nakka.com

**************************************************************************/

/**************************************************************************
	Include Files
**************************************************************************/

#include <windows.h>
#include <tchar.h>

#include "view.h"
#include "string.h"
#include "memory.h"


/**************************************************************************
	Define
**************************************************************************/

#define RESERVE_SZIE			4096
#define RESERVE_LINE			100


/**************************************************************************
	Global Variables
**************************************************************************/

/**************************************************************************
	Local Function Prototypes
**************************************************************************/

static BOOL IsLeadChar(TCHAR *c);
static void SetScrollBar(HWND hWnd, BUFFER *bf, int bottom);
static BOOL AllocLine(BUFFER *bf);
static BOOL CountLine(HWND hWnd, BUFFER *bf, TCHAR *st);
static void ReCountLine(BUFFER *bf, int ln);
static BOOL AddString(HWND hWnd, BUFFER *bf, TCHAR *str);
static void DrawLine(HWND hWnd, HDC mdc, BUFFER *bf, int i);


/******************************************************************************

	IsLeadChar

	Sp1oCgڂǂ`FbN

******************************************************************************/

static BOOL IsLeadChar(TCHAR *c)
{
	return ((WideCharToMultiByte(CP_ACP, 0, c, 1, NULL, 0, NULL, NULL) == 1) ? FALSE : TRUE);
}


/******************************************************************************

	Text2Clipboard

	Nbv{[hɐݒ

******************************************************************************/

static BOOL Text2Clipboard(HWND hWnd, TCHAR *st, TCHAR *en)
{
	HANDLE hMem;
	TCHAR *buf;

	//Nbv{[hJ
	if(OpenClipboard(hWnd) == FALSE){
		return FALSE;
	}
	if(EmptyClipboard() == FALSE){
		CloseClipboard();
		return FALSE;
	}

	if((hMem = LocalAlloc(LHND, sizeof(TCHAR) * (en - st + 1))) == NULL){
		CloseClipboard();
		return FALSE;
	}
	if((buf = LocalLock(hMem)) == NULL){
		LocalFree(hMem);
		CloseClipboard();
		return FALSE;
	}
	TStrCpyN(buf, st, en - st + 1);
	LocalUnlock(hMem);

#ifdef UNICODE
	SetClipboardData(CF_UNICODETEXT, hMem);
#else
	SetClipboardData(CF_TEXT, hMem);
#endif

	CloseClipboard();
	return TRUE;
}

/******************************************************************************

	SetScrollBar

	XN[o[̐ݒ

******************************************************************************/

static void SetScrollBar(HWND hWnd, BUFFER *bf, int bottom)
{
#define DISABLE_SCROLL_RANGE	32768
	SCROLLINFO si;

	ZeroMemory(&si, sizeof(SCROLLINFO));
	si.cbSize = sizeof(SCROLLINFO);
	si.fMask = SIF_RANGE | SIF_PAGE;

	if((bottom / bf->FontHeight) < bf->maxline){
		bf->max = bf->maxline - (bottom / bf->FontHeight);
		si.nPage = bottom / bf->FontHeight;
		si.nMax = bf->maxline - 1;
	}else{
		bf->max = 0;
		si.nPage = DISABLE_SCROLL_RANGE - 1;
		si.nMax = DISABLE_SCROLL_RANGE;
	}

	SetScrollInfo(hWnd, SB_VERT, &si, TRUE);
}


/******************************************************************************

	AllocLine

	s̊m

******************************************************************************/

static BOOL AllocLine(BUFFER *bf)
{
	int *mi;

	bf->linesize += RESERVE_LINE;
	mi = MemAlloc(sizeof(int) * bf->linesize);
	if(mi == NULL){
		return FALSE;
	}
	CopyMemory(mi, bf->line, sizeof(int) * bf->maxline);
	MemFree(&bf->line);
	bf->line = mi;

	mi = MemAlloc(sizeof(int) * bf->linesize);
	if(mi == NULL){
		return FALSE;
	}
	CopyMemory(mi, bf->linelen, sizeof(int) * bf->maxline);
	MemFree(&bf->linelen);
	bf->linelen = mi;
	return TRUE;
}


/******************************************************************************

	CountLine

	s̐ݒ

******************************************************************************/

static BOOL CountLine(HWND hWnd, BUFFER *bf, TCHAR *st)
{
	RECT rect;
	SIZE sz;
	TCHAR *p;
	int line_size;
	int len = 0;
	int cnt = 0;
	int offset = bf->LeftMargin;
	int width;
	int clen;

	GetClientRect(hWnd, &rect);
	line_size = rect.right - bf->RightMargin;

	for(p = st; *p != TEXT('\0'); p++){
		if(*p == TEXT('\r')){
		}else if(*p == TEXT('\n')){
			//new line
			*(bf->line + bf->maxline) = p + 1 - bf->buf;
			*(bf->linelen + bf->maxline - 1) = len;
			bf->maxline++;
			if(bf->maxline >= bf->linesize){
				if(AllocLine(bf) == FALSE){
					return FALSE;
				}
			}
			cnt = len = 0;
			offset = bf->LeftMargin;
		}else if(*p == TEXT('\t')){
			//tab
			offset += (bf->TabStop - (cnt % bf->TabStop)) * bf->CharWidth;
			cnt += bf->TabStop - (cnt % bf->TabStop);
			len++;
			if(offset > line_size){
				*(bf->line + bf->maxline) = p + 1 - bf->buf;
				*(bf->linelen + bf->maxline - 1) = len;
				bf->maxline++;
				if(bf->maxline >= bf->linesize){
					if(AllocLine(bf) == FALSE){
						return FALSE;
					}
				}
				cnt = len = 0;
				offset = bf->LeftMargin;
			}
		}else{
			//char
#ifdef UNICODE
			clen = 0;
#else
			clen = (IsDBCSLeadByte((BYTE)*p) == TRUE) ? 1 : 0;
#endif
			GetTextExtentPoint32(bf->mdc, p, clen + 1, &sz);
			width = sz.cx;

			if(offset + width >= line_size){
				//new line
				*(bf->line + bf->maxline) = p - bf->buf;
				*(bf->linelen + bf->maxline - 1) = len;
				bf->maxline++;
				if(bf->maxline >= bf->linesize){
					if(AllocLine(bf) == FALSE){
						return FALSE;
					}
				}
				cnt = len = clen + 1;
				offset = bf->LeftMargin + width;
			}else{
				len += clen + 1;
				cnt += clen + 1;
				offset += width;
			}
			p += clen;
		}
	}
	*(bf->linelen + bf->maxline - 1) = len;

	SetScrollBar(hWnd, bf, rect.bottom);
	return TRUE;
}


/******************************************************************************

	ReCountLine

	s̍Đݒ

******************************************************************************/

static void ReCountLine(BUFFER *bf, int ln)
{
	int i, j;

	i = *(bf->line + ln);

	for(j = ln; j < bf->maxline; j++){
		*(bf->line + j) -= i;
	}
	CopyMemory(bf->line, bf->line + ln, sizeof(int) * (bf->maxline - ln));
	CopyMemory(bf->linelen, bf->linelen + ln, sizeof(int) * (bf->maxline - ln));
	bf->maxline -= ln;
	bf->max -= ln;
	if(bf->max < 0){
		bf->max = 0;
	}
}


/******************************************************************************

	AddString

	̒ǉ

******************************************************************************/

static BOOL AddString(HWND hWnd, BUFFER *bf, TCHAR *str)
{
	RECT rect;
	TCHAR *p, *r, *s;
	int len, len2;
	int ln;
	int scroll_line, draw_line;
	BOOL scroll;

	len = bf->len;
	len2 = lstrlen(str);
	scroll = (bf->max == 0 || bf->pos == bf->max) ? TRUE : FALSE;
	p = str;
	if(bf->Limit <= 0){
		// no limit
		if(len + len2 + 1 > bf->size){
			bf->size = len + len2 + RESERVE_SZIE;

			s = MemAlloc(sizeof(TCHAR) * bf->size);
			if(s == NULL) return FALSE;
			r = TStrCpy(s, bf->buf);
			bf->sp = s + (bf->sp - bf->buf);
			bf->ep = s + (bf->ep - bf->buf);
			MemFree(&bf->buf);
			bf->buf = s;
		}else{
			r = bf->buf + len;
		}

	}else if(len + len2 >= bf->Limit){
		if(len2 > bf->Limit){
			p = str + len2 - bf->Limit + 1;
			r = bf->buf;
			bf->len = 0;
			bf->maxline = 1;
			*bf->line = 0;
		}else{
			r = bf->buf + len - bf->Limit + len2 + 1;
			for(; *r != TEXT('\0') && *r != TEXT('\n'); r++);
			for(; *r == TEXT('\r') || *r == TEXT('\n'); r++);
			for(ln = 0; ln < bf->maxline - 1 && *(bf->line + ln + 1) + bf->buf <= r; ln++);

			ReCountLine(bf, ln);

			bf->sp = bf->sp - (r - bf->buf);
			bf->ep = bf->ep - (r - bf->buf);
			if(bf->sp < bf->buf) bf->sp = bf->buf;
			if(bf->ep < bf->buf) bf->ep = bf->buf;

			bf->len -= (r - bf->buf);
			r = TStrCpy(bf->buf, r);
		}
	}else{
		r = bf->buf + len;
	}

	//copy
	s = r;
	while(*p != TEXT('\0')){
		if(*p == 0x08){
			if(r > bf->buf && *(r - 1) != TEXT('\n')){
				if(IsLeadChar(r - 1) == TRUE){
					*(r - 1) = TEXT(' ');
				}else{
					r--;
				}
			}
			p++;
			continue;
		}
		*(r++) = *(p++);
	}
	*r = TEXT('\0');
	bf->len = bf->len + (r - s);

	scroll_line = bf->max;
	draw_line = bf->maxline;
	//s̐ݒ
	if(CountLine(hWnd, bf, bf->buf + *(bf->line + bf->maxline - 1)) == FALSE){
		return FALSE;
	}
	scroll_line -= bf->max;
	draw_line -= bf->maxline;

	if(scroll == TRUE){
		GetClientRect(hWnd, &rect);
		if(bf->max > 0){
			//ǉXN[
			if(scroll_line != 0){
				ScrollWindowEx(hWnd, 0, scroll_line * bf->FontHeight, NULL, &rect, NULL, NULL, SW_INVALIDATE | SW_ERASE);
			}
			rect.top = ((rect.bottom / bf->FontHeight) + draw_line) * bf->FontHeight - bf->FontHeight;
			InvalidateRect(hWnd, &rect, FALSE);
		}else{
			InvalidateRect(hWnd, NULL, FALSE);
		}
		bf->pos = bf->max;
		SetScrollPos(hWnd, SB_VERT, bf->pos, TRUE);
		UpdateWindow(hWnd);

	}else if(bf->Limit > 0 && len + len2 > bf->Limit){
		InvalidateRect(hWnd, NULL, FALSE);
		UpdateWindow(hWnd);
	}
	return TRUE;
}


/******************************************************************************

	DrawLine

	1s`

******************************************************************************/

static void DrawLine(HWND hWnd, HDC mdc, BUFFER *bf, int i)
{
	RECT drect;
	SIZE sz;
	TCHAR *p, *s;
	int offset;
	int height;
	int csize;
	int cnt = 0;
	int j, tab;
	BOOL sel = FALSE;

	s = p = *(bf->line + i) + bf->buf;
	offset = bf->LeftMargin;
	height = (i - bf->pos) * bf->FontHeight + (bf->Spacing / 2);
	drect.top = (i - bf->pos) * bf->FontHeight;
	drect.bottom = drect.top + bf->FontHeight;

	if(bf->sp != bf->ep && ((p >= bf->sp && p < bf->ep) || (p >= bf->ep && p < bf->sp))){
		//I
		SetTextColor(mdc, GetSysColor(COLOR_HIGHLIGHTTEXT));
		SetBkColor(mdc, GetSysColor(COLOR_HIGHLIGHT));
		sel = TRUE;
	}else{
		//I
		SetTextColor(bf->mdc, GetSysColor(COLOR_WINDOWTEXT));
		SetBkColor(bf->mdc, GetSysColor(COLOR_WINDOW));
		sel = FALSE;
	}
	for(j = 0; j < *(bf->linelen + i); j++, p++){
		if(bf->sp != bf->ep && ((p >= bf->sp && p < bf->ep) || (p >= bf->ep && p < bf->sp))){
			if(sel == FALSE){
				GetTextExtentPoint32(mdc, s, p - s, &sz);
				csize = sz.cx;

				drect.left = offset;
				drect.right = offset + csize;
				ExtTextOut(mdc, offset, height, ETO_OPAQUE, &drect, s, p - s, NULL);
				offset += csize;
				s = p;

				//I
				SetTextColor(mdc, GetSysColor(COLOR_HIGHLIGHTTEXT));
				SetBkColor(mdc, GetSysColor(COLOR_HIGHLIGHT));
				sel = TRUE;
			}
		}else if(sel == TRUE){
			GetTextExtentPoint32(mdc, s, p - s, &sz);
			csize = sz.cx;

			drect.left = offset;
			drect.right = offset + csize;
			ExtTextOut(mdc, offset, height, ETO_OPAQUE, &drect, s, p - s, NULL);
			offset += csize;
			s = p;

			//I
			SetTextColor(bf->mdc, GetSysColor(COLOR_WINDOWTEXT));
			SetBkColor(bf->mdc, GetSysColor(COLOR_WINDOW));
			sel = FALSE;
		}
		if(*p == TEXT('\t')){
			RECT trect;
			HBRUSH hBrush;

			GetTextExtentPoint32(mdc, s, p - s, &sz);
			csize = sz.cx;

			drect.left = offset;
			drect.right = offset + csize;
			ExtTextOut(mdc, offset, height, ETO_OPAQUE, &drect, s, p - s, NULL);
			offset += csize;

			//tab
			tab = bf->TabStop - (cnt % bf->TabStop);

			trect.left = offset;
			trect.right = trect.left + tab * bf->CharWidth;
			trect.top = height - (bf->Spacing / 2);
			trect.bottom = trect.top + bf->FontHeight;

			hBrush = CreateSolidBrush(GetBkColor(mdc));
			FillRect(mdc, &trect, hBrush);
			DeleteObject(hBrush);

			offset += tab * bf->CharWidth;
			cnt += tab;
			s = p + 1;
			continue;
		}
#ifndef UNICODE
		if(IsDBCSLeadByte((BYTE)*p) == TRUE){
			p++;
			j++;
			cnt++;
		}
#endif
		cnt++;
	}
	GetTextExtentPoint32(mdc, s, p - s, &sz);
	drect.left = offset;
	drect.right = offset + sz.cx;
	ExtTextOut(mdc, offset, height, ETO_OPAQUE, &drect, s, p - s, NULL);

	if(*p == TEXT('\0')){
		bf->cx = offset + sz.cx;
	}
}


/******************************************************************************

	Point2Caret

	WLbg̈ʒu擾

******************************************************************************/

static TCHAR *Point2Caret(HWND hWnd, BUFFER *bf, int x, int y)
{
	RECT rect;
	SIZE sz;
	TCHAR *p, *r;
	int cnt;
	int offset;
	int clen;
	int ci;
	int i, j;

	GetClientRect(hWnd, &rect);
	ci = i = bf->pos + (y / bf->FontHeight);
	if(i < 0){
		return bf->buf;
	}else if(i >= bf->maxline){
		return bf->buf + bf->len;
	}
	for(; i > 0; i--){
		//_s̐擪Ɉړ
		p = *(bf->line + i) + bf->buf - 1;
		if(*p == '\r' || *p == '\n'){
			break;
		}
	}

	for(; i <= ci; i++){
		cnt = j = 0;
		offset = bf->LeftMargin - (bf->CharWidth / 2);
		for(r = p = *(bf->line + i) + bf->buf; j <= *(bf->linelen + i); p++, j++){
			if(i == ci && offset > x){
				break;
			}
			r = p;
			if(*p == TEXT('\t')){
				offset += (bf->TabStop - (cnt % bf->TabStop)) * bf->CharWidth;
				cnt += bf->TabStop - (cnt % bf->TabStop);
				continue;
			}
#ifdef UNICODE
			clen = 0;
#else
			clen = (IsDBCSLeadByte((BYTE)*p) == TRUE && *(p + 1) != TEXT('\0')) ? 1 : 0;
#endif
			GetTextExtentPoint32(bf->mdc, p, clen + 1, &sz);
			offset += sz.cx;

			p += clen;
			j += clen;
			cnt += clen + 1;
		}
	}
	return r;
}


/******************************************************************************

	GetCaretToken

	Lbgʒũg[N擾

******************************************************************************/

static void GetCaretToken(BUFFER *bf, TCHAR **st, TCHAR **en)
{
	TCHAR *p, *r;
	int i;

	for(i = 0; i < bf->maxline - 1 && *(bf->line + i + 1) + bf->buf <= bf->sp; i++);
	for(; i > 0; i--){
		//_s̐擪Ɉړ
		p = *(bf->line + i) + bf->buf - 1;
		if(*p == '\r' || *p == '\n'){
			break;
		}
	}

	p = *(bf->line + i) + bf->buf;
	while(*p != TEXT('\0') && *p != TEXT('\r') && *p != TEXT('\n')){
		if(IsLeadChar(p) == TRUE){
			r = p;
			while(IsLeadChar(p) == TRUE){
				p++;
			}

		}else if((*p >= TEXT('A') && *p <= TEXT('Z')) ||
			(*p >= TEXT('a') && *p <= TEXT('z')) ||
			(*p >= TEXT('0') && *p <= TEXT('9')) ||
			*p == TEXT('_')){
			r = p;
			while((*p >= TEXT('A') && *p <= TEXT('Z')) ||
				(*p >= TEXT('a') && *p <= TEXT('z')) ||
				(*p >= TEXT('0') && *p <= TEXT('9')) ||
				*p == TEXT('_')){
				p++;
			}

		}else{
			r = p++;
		}
		if(p > bf->sp){
			*st = r;
			*en = p;
			break;
		}
	}
}


/******************************************************************************

	ViewProc

	VIEWEBhEvV[W

******************************************************************************/

LRESULT CALLBACK ViewProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	BUFFER *bf;
	HDC hDC;
	RECT rect;
	TCHAR *p, *r;
	int i;

	switch(message)
	{
	case WM_CREATE:
		bf = MemAlloc(sizeof(BUFFER));
		if(bf == NULL) return -1;

		ZeroMemory(bf, sizeof(BUFFER));
		bf->Limit = 0;
		bf->TabStop = 8;
		bf->LeftMargin = 1;
		bf->RightMargin = 1;
		bf->Spacing = 2;

		bf->size = (bf->Limit <= 0) ? RESERVE_SZIE : bf->Limit;
		bf->linesize = RESERVE_LINE;
		bf->maxline = 1;

		//buf init
		bf->buf = MemAlloc(sizeof(TCHAR) * bf->size);
		if(bf->buf == NULL) return -1;
		*bf->buf = TEXT('\0');
		bf->len = 0;

		bf->InputBuf = MemAlloc(sizeof(TCHAR) * INPUT_SIZE);
		if(bf->InputBuf == NULL) return -1;
		*bf->InputBuf = TEXT('\0');

		//line init
		bf->line = MemAlloc(sizeof(int) * bf->linesize);
		if(bf->line == NULL) return -1;
		ZeroMemory(bf->line, sizeof(int) * bf->linesize);
		bf->linelen = MemAlloc(sizeof(int) * bf->linesize);
		if(bf->linelen == NULL) return -1;
		ZeroMemory(bf->linelen, sizeof(int) * bf->linesize);

		hDC = GetDC(hWnd);
		GetClientRect(hWnd, &rect);
		bf->mdc = CreateCompatibleDC(hDC);
		bf->hBmp = CreateCompatibleBitmap(hDC, rect.right, rect.bottom);
		bf->hRetBmp = SelectObject(bf->mdc, bf->hBmp);
		ReleaseDC(hWnd, hDC);

		SetTextColor(bf->mdc, GetSysColor(COLOR_WINDOWTEXT));
		SetBkColor(bf->mdc, GetSysColor(COLOR_WINDOW));

		bf->hBkBrush = CreateSolidBrush(GetSysColor(COLOR_WINDOW));

		bf->echo = TRUE;

		//buffer info to window long
		SetWindowLong(hWnd, GWL_USERDATA, (LPARAM)bf);

		SendMessage(hWnd, WM_REFLECT, 0, 0);
		break;

	case WM_DESTROY:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf != NULL){
			SetWindowLong(hWnd, GWL_USERDATA, (LPARAM)0);

			SelectObject(bf->mdc, bf->hRetBmp);
			DeleteObject(bf->hBmp);
			DeleteDC(bf->mdc);

			DeleteObject(bf->hBkBrush);

			MemFree(&bf->buf);
			MemFree(&bf->InputBuf);
			MemFree(&bf->line);
			MemFree(&bf->linelen);
			MemFree(&bf);
		}

		if(GetFocus() == hWnd){
			DestroyCaret();
		}
		return DefWindowProc(hWnd, message, wParam, lParam);

	case WM_SETFOCUS:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;
		CreateCaret(hWnd, 0, 0, bf->FontHeight);
		InvalidateRect(hWnd, NULL, FALSE);

		ShowCaret(hWnd);
		break;

	case WM_KILLFOCUS:
		HideCaret(hWnd);
		DestroyCaret();
		break;

	case WM_GETDLGCODE:
		return DLGC_WANTALLKEYS;

	case WM_VSCROLL:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;
		GetClientRect(hWnd, &rect);
		i = bf->pos;
		switch((int)LOWORD(wParam))
		{
		case SB_TOP:
			bf->pos = 0;
			break;

		case SB_BOTTOM:
			bf->pos = bf->max;
			break;

		case SB_LINEDOWN:
			bf->pos = (bf->pos < bf->max) ? bf->pos + 1 : bf->max;
			break;

		case SB_LINEUP:
			bf->pos = (bf->pos > 0) ? bf->pos - 1 : 0;
			break;

		case SB_PAGEDOWN:
			bf->pos = (bf->pos + (rect.bottom / bf->FontHeight) < bf->max) ?
				bf->pos + (rect.bottom / bf->FontHeight) : bf->max;
			break;

		case SB_PAGEUP:
			bf->pos = (bf->pos - (rect.bottom / bf->FontHeight) > 0) ?
				bf->pos - (rect.bottom / bf->FontHeight) : 0;
			break;

		case SB_THUMBPOSITION:
		case SB_THUMBTRACK:
			if((short)HIWORD(wParam) > bf->max){
				bf->pos = bf->max;
			}else if((short)HIWORD(wParam) < 0){
				bf->pos = 0;
			}else{
				bf->pos = HIWORD(wParam);
			}
			break;
		}

		SetScrollPos(hWnd, SB_VERT, bf->pos, TRUE);

		ScrollWindowEx(hWnd, 0, (i - bf->pos) * bf->FontHeight, NULL, &rect, NULL, NULL, SW_INVALIDATE | SW_ERASE);
		UpdateWindow(hWnd);
		break;

	case WM_KEYDOWN:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;

		switch(wParam)
		{
		case VK_HOME:
			SendMessage(hWnd, WM_VSCROLL, SB_TOP, 0);
			break;

		case VK_END:
			SendMessage(hWnd, WM_VSCROLL, SB_BOTTOM, 0);
			break;

		case VK_PRIOR:
			SendMessage(hWnd, WM_VSCROLL, SB_PAGEUP, 0);
			break;

		case VK_NEXT:
			SendMessage(hWnd, WM_VSCROLL, SB_PAGEDOWN, 0);
			break;

		case VK_UP:
			SendMessage(hWnd, WM_VSCROLL, SB_LINEUP, 0);
			break;

		case VK_DOWN:
			SendMessage(hWnd, WM_VSCROLL, SB_LINEDOWN, 0);
			break;
		}
		break;

	case WM_LBUTTONDOWN:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;

		SetCapture(hWnd);
		SetFocus(hWnd);

		bf->mousedown = TRUE;
		bf->sp = bf->ep = Point2Caret(hWnd, bf, (short)LOWORD(lParam), (short)HIWORD(lParam));

		InvalidateRect(hWnd, NULL, FALSE);
		UpdateWindow(hWnd);
		break;

	case WM_MOUSEMOVE:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;
		if(!(wParam & MK_LBUTTON) || bf->mousedown == FALSE) break;

		GetClientRect(hWnd, &rect);
		i = bf->pos + ((short)HIWORD(lParam) / bf->FontHeight);
		if(i < bf->pos){
			if(i < 0) i = 0;
			ScrollWindowEx(hWnd, 0,
				(bf->pos - i) * bf->FontHeight,
				NULL, &rect, NULL, NULL, SW_INVALIDATE | SW_ERASE);
			bf->pos = i;
			SetScrollPos(hWnd, SB_VERT, bf->pos, TRUE);
		}else if(i >= bf->pos + (rect.bottom / bf->FontHeight) && bf->max != 0){
			if(i >= bf->maxline - 1) i = bf->maxline - 1;
			ScrollWindowEx(hWnd, 0,
				(bf->pos - (i - (rect.bottom / bf->FontHeight) + 1)) * bf->FontHeight,
				NULL, &rect, NULL, NULL, SW_INVALIDATE | SW_ERASE);
			bf->pos = i - (rect.bottom / bf->FontHeight) + 1;
			bf->pos = (bf->pos > 0 && bf->pos < bf->max) ? bf->pos : bf->max;
			SetScrollPos(hWnd, SB_VERT, bf->pos, TRUE);
		}
		bf->ep = Point2Caret(hWnd, bf, (short)LOWORD(lParam), (short)HIWORD(lParam));

		InvalidateRect(hWnd, NULL, FALSE);
		UpdateWindow(hWnd);
		break;

	case WM_LBUTTONUP:
		ReleaseCapture();

		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;
		if(bf->mousedown == FALSE) break;

		SetFocus(hWnd);

		bf->mousedown = FALSE;
		bf->ep = Point2Caret(hWnd, bf, (short)LOWORD(lParam), (short)HIWORD(lParam));

		InvalidateRect(hWnd, NULL, FALSE);
		UpdateWindow(hWnd);
		break;

	case WM_LBUTTONDBLCLK:
		if(!(wParam & MK_LBUTTON)) break;
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;
		//select token
		GetCaretToken(bf, &bf->sp, &bf->ep);
		InvalidateRect(hWnd, NULL, FALSE);
		UpdateWindow(hWnd);
		break;

	case WM_PAINT:
		{
			PAINTSTRUCT ps;

			bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
			if(bf == NULL) break;
			hDC = BeginPaint(hWnd, &ps);

			FillRect(bf->mdc, &ps.rcPaint, bf->hBkBrush);

			i = bf->pos + (ps.rcPaint.top / bf->FontHeight);
			if(i >= bf->maxline) i = bf->maxline - 1;
			for(; i < bf->maxline && i < bf->pos + (ps.rcPaint.bottom / bf->FontHeight) + 1; i++){
				//draw line
				DrawLine(hWnd, bf->mdc, bf, i);
			}

			if(GetFocus() == hWnd){
				SetCaretPos(bf->cx, ((bf->maxline - 1) - bf->pos) * bf->FontHeight);
			}

			BitBlt(hDC, ps.rcPaint.left, ps.rcPaint.top, ps.rcPaint.right, ps.rcPaint.bottom,
				bf->mdc, ps.rcPaint.left, ps.rcPaint.top, SRCCOPY);

			EndPaint(hWnd, &ps);
		}
		break;

	case WM_ERASEBKGND:
		return 1;

	case WM_SIZE:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;

		i = (bf->max == 0 || bf->pos == bf->max) ? 1 : 0;
		GetClientRect(hWnd, &rect);

		if(bf->width != rect.right){
			bf->width = rect.right;

			//clear line info
			MemFree(&bf->line);
			MemFree(&bf->linelen);

			bf->linesize = RESERVE_LINE;
			bf->line = MemAlloc(sizeof(int) * bf->linesize);
			if(bf->line == NULL) break;
			ZeroMemory(bf->line, sizeof(int) * bf->linesize);
			bf->linelen = MemAlloc(sizeof(int) * bf->linesize);
			if(bf->linelen == NULL) break;
			ZeroMemory(bf->linelen, sizeof(int) * bf->linesize);

			//set line info
			bf->maxline = 1;
			CountLine(hWnd, bf, bf->buf);
		}else{
			SetScrollBar(hWnd, bf, rect.bottom);
		}

		if(i == 1 || bf->pos > bf->max){
			bf->pos = bf->max;
			SetScrollPos(hWnd, SB_VERT, bf->pos, TRUE);
		}

		hDC = GetDC(hWnd);
		SelectObject(bf->mdc, bf->hRetBmp);
		DeleteObject(bf->hBmp);
		bf->hBmp = CreateCompatibleBitmap(hDC, rect.right, rect.bottom);
		bf->hRetBmp = SelectObject(bf->mdc, bf->hBmp);
		ReleaseDC(hWnd, hDC);

		InvalidateRect(hWnd, NULL, FALSE);
		break;

	case WM_SETTEXT:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;
		if(bf->Limit <= 0){
			MemFree(&bf->buf);
			MemFree(&bf->line);
			MemFree(&bf->linelen);

			bf->size = RESERVE_SZIE;
			bf->linesize = RESERVE_LINE;

			//buf init
			bf->buf = MemAlloc(sizeof(TCHAR) * bf->size);
			if(bf->buf == NULL) break;

			//line init
			bf->line = MemAlloc(sizeof(int) * bf->linesize);
			if(bf->line == NULL) break;
			bf->linelen = MemAlloc(sizeof(int) * bf->linesize);
			if(bf->linelen == NULL) break;
		}
		*bf->buf = TEXT('\0');
		bf->len = 0;
		*bf->line = 0;
		bf->maxline = 1;
		bf->pos = bf->max = 0;
		return AddString(hWnd, bf, (TCHAR *)lParam);

	case WM_ADDTEXT:
		if(lParam == 0) break;
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;
		return AddString(hWnd, bf, (TCHAR *)lParam);

	case WM_GETBUFFERINFO:
		*((BUFFER **)lParam) = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		break;

	case WM_REFLECT:
		{
			TEXTMETRIC tm;

			bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
			if(bf == NULL) break;

			//Metrics
			GetTextMetrics(bf->mdc, &tm);
			bf->FontHeight = tm.tmHeight + bf->Spacing;
			bf->CharWidth = tm.tmAveCharWidth;

			if(GetFocus() == hWnd){
				DestroyCaret();
				CreateCaret(hWnd, 0, 0, bf->FontHeight);
				ShowCaret(hWnd);
			}
			bf->width = 0;
			SendMessage(hWnd, WM_SIZE, 0, 0);
		}
		break;

	case WM_INITVIEW:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;
		*bf->InputBuf = TEXT('\0');
		bf->echo = TRUE;
		break;

	case WM_GETCHAR:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;
		bf->tid = lParam;
		bf->lineFlag = FALSE;
		bf->lineBuf = NULL;

		if(*bf->InputBuf != TEXT('\0')){
			p = AllocCopy(bf->InputBuf);
			*bf->InputBuf = TEXT('\0');
			for(i = 0; (unsigned int)i < lstrlen(p); i++){
				SendMessage(hWnd, WM_CHAR, *(p + i), 0);
			}
			MemFree(&p);
		}
		break;

	case WM_GETS:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;
		bf->tid = lParam;
		bf->lineFlag = TRUE;
		bf->lineBuf = NULL;

		if(*bf->InputBuf != TEXT('\0')){
			p = AllocCopy(bf->InputBuf);
			*bf->InputBuf = TEXT('\0');
			for(i = 0; (unsigned int)i < lstrlen(p); i++){
				SendMessage(hWnd, WM_CHAR, *(p + i), 0);
			}
			MemFree(&p);
		}
		break;

	case WM_SETECHO:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;
		bf->echo = lParam;
		break;

	case WM_CHAR:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;
		if(bf->tid == 0){
			i = lstrlen(bf->InputBuf);
			if(i >= INPUT_SIZE - 1){
				break;
			}
			*(bf->InputBuf + i) = wParam;
			*(bf->InputBuf + i + 1) = TEXT('\0');
			break;
		}
		SendMessage(hWnd, WM_VSCROLL, SB_BOTTOM, 0);
		if(wParam == VK_BACK && bf->lineFlag == TRUE){
			if(bf->lineBuf != NULL && lstrlen(bf->lineBuf) > 0 && bf->linelen[bf->maxline - 1] > 0){
				*(bf->lineBuf + lstrlen(bf->lineBuf) - 1) = TEXT('\0');
				*(bf->buf + bf->len - 1) = TEXT('\0');
				bf->len--;
				bf->linelen[bf->maxline - 1]--;
			}
			InvalidateRect(hWnd, NULL, FALSE);
			break;
		}

		p = MemAlloc(sizeof(TCHAR) * 2);
		p[0] = wParam;
		p[1] = TEXT('\0');

		if(p[0] == TEXT('\n')){
			p[0] = TEXT('\0');
		}else if(p[0] == TEXT('\r')){
			p[0] = TEXT('\n');
		}

		if(bf->echo == TRUE){
			if(p[0] != '\n'){
				SendMessage(hWnd, WM_ADDTEXT, 0, (LPARAM)p);
			}else{
				SendMessage(hWnd, WM_ADDTEXT, 0, (LPARAM)TEXT("\r\n"));
			}
		}

		if(bf->lineFlag == FALSE){
			PostThreadMessage((DWORD)bf->tid, TH_GETCHAR, 0, (LPARAM)p);
			bf->tid = 0;
		}else{
			if(p[0] == '\n'){
				PostThreadMessage((DWORD)bf->tid, TH_GETS, 0, (LPARAM)bf->lineBuf);
				bf->tid = 0;
				bf->lineBuf = NULL;
			}else{
				r = AllocJoin(bf->lineBuf, p);
				MemFree(&bf->lineBuf);
				bf->lineBuf = r;
			}
		}
		break;

	case WM_COPY:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;
		if(bf->sp != bf->ep){
			if(bf->sp > bf->ep){
				Text2Clipboard(hWnd, bf->ep, bf->sp);
			}else{
				Text2Clipboard(hWnd, bf->sp, bf->ep);
			}
		}
		break;

	case WM_PASTE:
		{
			UINT ClipFormat[1];
			HANDLE hMem;
			TCHAR *buf;

			ClipFormat[0] = CF_UNICODETEXT;
			if(GetPriorityClipboardFormat(ClipFormat, 1) != CF_UNICODETEXT){
				break;
			}
			if(OpenClipboard(hWnd) == FALSE){
				break;
			}
			hMem = GetClipboardData(CF_UNICODETEXT);
			if((buf = LocalLock(hMem)) == NULL){
				CloseClipboard();
				break;
			}
			for(p = buf; *p != TEXT('\0'); p++){
				SendMessage(hWnd, WM_CHAR, *p, 0);
			}
			LocalUnlock(hMem);
			CloseClipboard();
		}
		break;

	case EM_LIMITTEXT:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;
		if((int)wParam <= 0){
			bf->Limit = 0;
			break;
		}

		bf->size = bf->Limit = wParam;
		if(bf->len > bf->Limit){
			r = bf->buf + bf->len - bf->Limit + 1;
			for(; *r != TEXT('\0') && *r != TEXT('\n'); r++);
			for(; *r == TEXT('\r') || *r == TEXT('\n'); r++);
		}else{
			r = bf->buf;
		}

		p = MemAlloc(sizeof(TCHAR) * bf->size);
		lstrcpy(p, r);
		MemFree(&bf->buf);
		bf->buf = p;
		bf->len = lstrlen(p);

		MemFree(&bf->line);
		MemFree(&bf->linelen);

		//line init
		bf->linesize = RESERVE_LINE;
		bf->line = MemAlloc(sizeof(int) * bf->linesize);
		bf->linelen = MemAlloc(sizeof(int) * bf->linesize);
		*bf->line = 0;
		bf->maxline = 1;
		bf->max = 0;

		//set line
		CountLine(hWnd, bf, bf->buf);

		bf->pos = bf->max;

		SetScrollPos(hWnd, SB_VERT, bf->pos, TRUE);
		InvalidateRect(hWnd, NULL, FALSE);
		UpdateWindow(hWnd);
		break;

	case EM_SETSEL:
		bf = (BUFFER *)GetWindowLong(hWnd, GWL_USERDATA);
		if(bf == NULL) break;

		bf->sp = bf->buf + wParam;
		if((int)lParam == -1){
			bf->ep = bf->buf + bf->len;
		}else{
			bf->ep = bf->buf + lParam;
		}

		InvalidateRect(hWnd, NULL, FALSE);
		UpdateWindow(hWnd);
		break;

	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}


/******************************************************************************

	RegistViewWindow

	VIEWNX̓o^

******************************************************************************/

BOOL RegistViewWindow(HINSTANCE hInstance)
{
	WNDCLASS wc;

	wc.style = CS_DBLCLKS;
	wc.lpfnWndProc = (WNDPROC)ViewProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor(0, IDC_IBEAM);
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
	wc.lpszMenuName = 0;
	wc.lpszClassName = VIEW_WND_CLASS;
	return RegisterClass(&wc);
}
/* End of source */
